#!/usr/bin/env python
#
# __init__.py - OpenGL data and rendering for FSLeyes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package contains the OpenGL data and rendering stuff for *FSLeyes*.
On-screen and off-screen rendering is supported, and two OpenGL versions (1.4
and 2.1) are supported.  The contents of this package can be broadly
categorised into the following:

 - *Canvases*: A canvas is a thing that can be drawn on.

 - *Objects*:  An object is a thing which can be drawn on a canvas.


-----------
Quick start
-----------

::

    import fsleyes.gl                 as fslgl
    import fsleyes.gl.wxglslicecanvas as slicecanvas

    # This function will be called when
    # the GL context is ready to be used.
    def ready():

        # The fsleyes.gl package needs to do
        # some initialisation that can only
        # be performed once a GL context has
        # been created.
        fslgl.bootstrap()

        # Once a GL context has been created,
        # you can do stuff! The SliceCanvas
        # will take care of creating and
        # managing GLObjects for each overlay
        # in the overlay list.
        canvas = slicecanvas.WXGLSliceCanvas(parent, overlayList, displayCtx)

    # Create a GL context, and tell it to
    # call our function when it is ready.
    fslgl.getGLContext(ready=ready)

    # ...

    # When you're finished, call the shutdown
    # function to clear the context (only
    # necessary for on-screen rendering)
    fslgl.shutdown()


--------
Canvases
--------


A *canvas* is the destination for an OpenGL scene render. The following
canvases are defined in the ``gl`` package:

.. autosummary::
   :nosignatures:

   ~fsleyes.gl.slicecanvas.SliceCanvas
   ~fsleyes.gl.lightboxcanvas.LightBoxCanvas
   ~fsleyes.gl.scene3dcanvas.Scene3DCanvas
   ~fsleyes.gl.colourbarcanvas.ColourBarCanvas


These classes are not intended to be used directly. This is because the ``gl``
package has been written to support two primary use-cases:

  - *On-screen* display of a scene using a :class:`wx.glcanvas.GLCanvas`
    canvas.

  - *Off-screen* rendering of a scene to a file.


Because of this, the canvas classes listed above are not dependent upon the
OpenGL environment in which they are used (i.e. on-screen or off-screen).
Instead, two base classes are provided for each of the use-cases:


.. autosummary::
   :nosignatures:

   WXGLCanvasTarget
   OffScreenCanvasTarget


And the following sub-classes are defined, providing use-case specific
implementations for each of the available canvases:

.. autosummary::
   :nosignatures:

   ~fsleyes.gl.wxglslicecanvas.WXGLSliceCanvas
   ~fsleyes.gl.wxgllightboxcanvas.WXGLLightBoxCanvas
   ~fsleyes.gl.wxglscene3dcanvas.WXGLScene3DCanvas
   ~fsleyes.gl.wxglcolourbarcanvas.WXGLColourBarCanvas
   ~fsleyes.gl.offscreenslicecanvas.OffScreenSliceCanvas
   ~fsleyes.gl.offscreenlightboxcanvas.OffScreenLightBoxCanvas
   ~fsleyes.gl.offscreenscene3dcanvas.OffScreenScene3DCanvas
   ~fsleyes.gl.offscreencolourbarcanvas.OffScreenColourBarCanvas


The classes listed above are the ones which are intended to be instantiated
and used by application code.


--------------
``gl`` objects
--------------


With the exception of the :class:`.ColourBarCanvas`, everything that is drawn
on a canvas derives from the :class:`.GLObject` base class. A ``GLObject``
manages the underlying data structures, GL resources (e.g. shaders and
textures), and rendering routines required to draw an object, in 2D or 3D, to a
canvas.  The following ``GLObject`` sub-classes correspond to each of the
possible types (the :attr:`.Display.overlayType` property) that an overlay can
be displayed as:

.. autosummary::

   ~fsleyes.gl.glvolume.GLVolume
   ~fsleyes.gl.glrgbvolume.GLRGBVolume
   ~fsleyes.gl.glcomplex.GLComplex
   ~fsleyes.gl.glmask.GLMask
   ~fsleyes.gl.gllabel.GLLabel
   ~fsleyes.gl.gllinevector.GLLineVector
   ~fsleyes.gl.glrgbvector.GLRGBVector
   ~fsleyes.gl.glmesh.GLMesh
   ~fsleyes.gl.glmip.GLMIP
   ~fsleyes.gl.gltensor.GLTensor
   ~fsleyes.gl.glsh.GLSH


These objects are created and destroyed automatically by the canvas classes
instances, so application code does not need to worry about them too much.


===========
Annotations
===========


Canvases can be *annotated* in a few ways, by use of the :class:`.Annotations`
class. An ``Annotations`` object allows lines, rectangles, and other simple
shapes to be rendered on top of the ``GLObject`` renderings which represent
the overlays in the :class:`.OverlayList`.  The ``Annotations`` object for a
canvas instance can be accessed through its ``getAnnotations`` method.


---------------------------------
OpenGL versions and bootstrapping
---------------------------------


*FSLeyes* needs to be able to run in restricted environments, such as within a
VNC session, and over SSH. In such environments the available OpenGL version
could be quite old, so the ``gl`` package has been written to support an
environment as old as OpenGL 1.4.


The available OpenGL API version can only be determined once an OpenGL context
has been created, and a display is available for rendering. The package-level
:func:`getGLContext` function allows a context to be created.


The data structures and rendering logic for some ``GLObject`` classes differs
depending on the OpenGL version that is available. Therefore, the code for
these ``GLObject`` classes may be duplicated, with one version for OpenGL 1.4,
and another version for OpenGL 2.1.  The ``GLObject`` code which targets a
specific OpenGL version lives within either the :mod:`.gl14` or :mod:`.gl21`
sub-packages.


Because of this, the package-level :func:`bootstrap` function must be called
before any ``GLObject`` instances are created, but *after* a GL context has
been created.
"""


import os
import sys
import logging
import ctypes.util as ctypes_util
import platform

import fsl.utils.idle                     as idle
import fsl.version                        as fslversion
from   fsl.utils.platform import platform as fslplatform
import fsleyes_widgets                    as fwidgets


real_find_library = ctypes_util.find_library


def find_library(name):
    if platform.system() == 'Darwin' and name in ('OpenGL', 'GLUT'):
        return f'/System/Library/Frameworks/{name}.framework/{name}'
    else:
        return real_find_library(name)


ctypes_util.find_library = find_library


log = logging.getLogger(__name__)


import OpenGL  # noqa


# Make PyOpenGL throw an error, instead of implicitly
# converting, if we pass incorrect types to OpenGL functions.
OpenGL.ERROR_ON_COPY = True


# These flags should be set to True
# for development, False for production
OpenGL.ERROR_CHECKING = False
OpenGL.ERROR_LOGGING  = False


# If FULL_LOGGING is enabled,
# every GL call will be logged.
# OpenGL.FULL_LOGGING   = True


GL_VERSION = None
"""Set in :func:`bootstrap`. String containing the available "major.minor"
OpenGL version.
"""


GL_COMPATIBILITY = None
"""Set in :func:`bootstrap`. String containing the target "major.minor"
OpenGL compatibility version ("1.4" or "2.1").
"""


GL_RENDERER = None
"""Set in :func:`bootstrap`. Contains a description of the OpenGL renderer in
 use.
"""


def _selectPyOpenGLPlatform():
    """Pyopengl sometimes doesn't select a suitable platform, so in some
    circumstances we need to force things (but not if ``PYOPENGL_PLATFORM``
    is already set in the environment).
    """
    if 'PYOPENGL_PLATFORM' in os.environ:
        return

    # If no display, osmesa on all platforms
    if not fwidgets.canHaveGui():
        os.environ['PYOPENGL_PLATFORM'] = 'osmesa'

    # GTK3 versions of wxpython 4.1.1 and newer
    # default to using EGL for GL initialisation,
    # but pyopengl defaults to glx.
    elif fslplatform.os.lower() == 'linux':
        wxver  = fwidgets.wxVersion()
        wxplat = fwidgets.wxPlatform()

        if wxver is not None                               and \
           fslversion.compareVersions(wxver, '4.1.1') >= 0 and \
           wxplat == fwidgets.WX_GTK3:
            os.environ['PYOPENGL_PLATFORM'] = 'egl'


_selectPyOpenGLPlatform()


def glIsSoftwareRenderer():
    """Returns ``True`` if the OpenGL renderer appears to be software based,
    ``False`` otherwise, or ``None`` :func:`bootstrap` has not been called yet.

    .. note:: This check is based on heuristics, ans is not guaranteed to
              be correct.
    """
    if GL_RENDERER is None:
        return None

    # There doesn't seem to be any quantitative
    # method for determining whether we are using
    # software-based rendering, so a hack is
    # necessary.
    renderer = GL_RENDERER.lower()

    # "software" / "chromium" -> software renderer
    # But SVGA3D/llvmpipe are super fast, so if
    # we're using either of them, pretend that
    # we're on hardware
    sw     = any(('software' in renderer, 'chromium' in renderer))
    fastsw = any(('llvmpipe' in renderer, 'svga3d'   in renderer))

    return sw and (not fastsw)


def bootstrap(glVersion=None):
    """Imports modules appropriate to the specified OpenGL version.

    The available OpenGL API version can only be queried once an OpenGL
    context is created, and a canvas is available to draw on. This makes
    things a bit complicated, because it means that we are only able to
    choose how to draw things when we actually need to draw them.


    This function should be called after an OpenGL context has been created,
    and a canvas is available for drawing, but before any attempt to draw
    anything.  It will figure out which version-dependent package needs to be
    loaded, and will attach all of the modules contained in said package to
    the :mod:`~fsleyes.gl` package.  The version-independent modules may
    then simply access these version-dependent modules through this module.


    After the :func:`boostrap` function has been called, the following
    package-level attributes will be available on the ``gl`` package:


    ====================== ====================================================
    ``GL_COMPATIBILITY``   A string containing the target OpenGL version, in
                           the format ``major.minor``, e.g. ``2.1``.

    ``GL_VERSION``         A string containing the available OpenGL version.

    ``GL_RENDERER``        A string containing the name of the OpenGL renderer.

    ``glvolume_funcs``     The version-specific module containing functions for
                           rendering :class:`.GLVolume` instances.

    ``glrgbvolume_funcs``  The version-specific module containing functions for
                           rendering :class:`.GLRGBVolume` instances.

    ``glrgbvector_funcs``  The version-specific module containing functions for
                           rendering :class:`.GLRGBVector` instances.

    ``gllinevector_funcs`` The version-specific module containing functions for
                           rendering :class:`.GLLineVector` instances.

    ``glmesh_funcs``       The version-specific module containing functions for
                           rendering :class:`.GLMesh` instances.

    ``glmask_funcs``       The version-specific module containing functions for
                           rendering :class:`.GLMask` instances.

    ``gllabel_funcs``      The version-specific module containing functions for
                           rendering :class:`.GLLabel` instances.

    ``gltensor_funcs``     The version-specific module containing functions for
                           rendering :class:`.GLTensor` instances.

    ``glsh_funcs``         The version-specific module containing functions for
                           rendering :class:`.GLSH` instances.

    ``glmip_funcs``        The version-specific module containing functions for
                           rendering :class:`.GLMIP` instances.
    ====================== ====================================================


    :arg glVersion: A tuple containing the desired (major, minor) OpenGL API
                    version to use. If ``None``, the best possible API
                    version will be used.
    """

    import OpenGL.GL         as gl
    import OpenGL.extensions as glexts
    from . import               gl14
    from . import               gl21

    thismod = sys.modules[__name__]

    if hasattr(thismod, '_bootstrapped'):
        return

    if glVersion is None:
        glver = gl.glGetString(gl.GL_VERSION).decode('latin1').split()[0]
        major, minor = [int(v) for v in glver.split('.')][:2]
    else:
        major, minor = glVersion

    # glVersion contains the actual GL version
    # verstr contains the target compatibility
    # GL version
    glVersion = major + minor / 10.0
    glpkg     = None
    if glVersion >= 2.1:
        verstr = '2.1'
        glpkg  = gl21
    elif glVersion >= 1.4:
        verstr = '1.4'
        glpkg  = gl14
    else: raise RuntimeError('OpenGL 1.4 or newer is required '
                             '(detected version: {:0.1f})'.format(glVersion))

    # The gl21 implementation depends on a
    # few extensions - if they're not present,
    # fall back to the gl14 implementation
    if glpkg == gl21:

        # List any GL21 extensions here
        exts = ['GL_EXT_framebuffer_object',
                'GL_ARB_instanced_arrays',
                'GL_ARB_draw_instanced']

        if not all(map(glexts.hasExtension, exts)):
            log.warning('One of these OpenGL extensions is '
                        'not available: [{}]. Falling back '
                        'to an older OpenGL implementation.'
                        .format(', '.join(exts)))
            verstr = '1.4'
            glpkg  = gl14

    # If using GL14, and the ARB_vertex_program
    # and ARB_fragment_program extensions are
    # not present, we're screwed.
    if glpkg == gl14:

        exts = ['GL_EXT_framebuffer_object',
                'GL_ARB_vertex_program',
                'GL_ARB_fragment_program',
                'GL_ARB_texture_non_power_of_two']

        if not all(map(glexts.hasExtension, exts)):
            raise RuntimeError('One of these OpenGL extensions is '
                               'not available: [{}]. This software '
                               'cannot run on the available graphics '
                               'hardware.'.format(', '.join(exts)))

        # Tensor/SH/MIP overlays are not available in GL14
        import fsleyes.displaycontext as dc
        dc.ALL_OVERLAY_TYPES            .remove('tensor')
        dc.ALL_OVERLAY_TYPES            .remove('sh')
        dc.ALL_OVERLAY_TYPES            .remove('mip')
        dc.OVERLAY_TYPES['DTIFitTensor'].remove('tensor')
        dc.OVERLAY_TYPES['Image']       .remove('sh')
        dc.OVERLAY_TYPES['Image']       .remove('tensor')
        dc.OVERLAY_TYPES['Image']       .remove('mip')

    renderer = gl.glGetString(gl.GL_RENDERER).decode('latin1')
    log.debug('Using OpenGL {} implementation with renderer {}'.format(
        verstr, renderer))

    thismod.GL_VERSION         = str(glVersion)
    thismod.GL_COMPATIBILITY   = verstr
    thismod.GL_RENDERER        = renderer
    thismod.glvolume_funcs     = glpkg.glvolume_funcs
    thismod.glrgbvolume_funcs  = glpkg.glrgbvolume_funcs
    thismod.glrgbvector_funcs  = glpkg.glrgbvector_funcs
    thismod.gllinevector_funcs = glpkg.gllinevector_funcs
    thismod.glmask_funcs       = glpkg.glmask_funcs
    thismod.glmesh_funcs       = glpkg.glmesh_funcs
    thismod.gllabel_funcs      = glpkg.gllabel_funcs
    thismod.gltensor_funcs     = glpkg.gltensor_funcs
    thismod.glsh_funcs         = glpkg.glsh_funcs
    thismod.glmip_funcs        = glpkg.glmip_funcs
    thismod._bootstrapped      = True

    # If we're using a software based renderer,
    # reduce the default performance settings
    if glIsSoftwareRenderer():

        log.debug('Software-based rendering detected - '
                  'lowering default performance settings.')

        import fsleyes.displaycontext as dc
        dc.SceneOpts.performance.setAttribute(None, 'default', 1)


def getGLContext(**kwargs):
    """Create and return a GL context object for on- or off-screen OpenGL
    rendering.

    If a context object has already been created, it is returned.
    Otherwise, one is created and returned.

    See the :class:`GLContext` class for details on the arguments.

    .. warning:: Use the ``ready`` argument to
                 :meth:`GLContext.__init__`, and don't call :func:`bootstrap`
                 until it has been called!
    """

    thismod = sys.modules[__name__]

    # A context has already been created
    if hasattr(thismod, '_glContext'):

        # If a callback was provided,
        # make sure it gets called.
        callback = kwargs.pop('ready', None)
        if callback is not None:
            idle.idle(callback)

        return thismod._glContext

    thismod._glContext = GLContext(**kwargs)

    return thismod._glContext


def shutdown():
    """Must be called when the GL rendering context is no longer needed.
    Destroys the context object, and resources associated with it.

    Does not need to be called for off-screen rendering.
    """

    thismod = sys.modules[__name__]
    context = getattr(thismod, '_glContext', None)
    if context is not None:
        context.destroy()
        delattr(thismod, '_glContext')


class GLContext:
    """The ``GLContext`` class manages the creation of, and access to, an
    OpenGL context. This class abstracts away the differences between
    creation of on-screen and off-screen rendering contexts.
    It contains two methods:

      - :meth:`setTarget`, which may be used to set a
        :class:`.WXGLCanvasTarget` or an :class:`OffScreenCanvasTarget` as the
        GL rendering target.
      - :meth:`destroy`, which must be called when the context is no longer
        needed.


    On-screen rendering is performed via the ``wx.GLCanvas.GLContext``
    context, whereas off-screen rendering is performed  via
    ``OpenGL.raw.osmesa.mesa`` (OSMesa is assumed to be available).


    If it is possible to do so, a ``wx.glcanvas.GLContext`` will be created,
    even if an off-screen context has been requested. This is because
    using the native graphics card is nearly always preferable to using
    OSMesa.


    *Creating an on-screen GL context*


    A ``wx.glcanvas.GLContext`` may only be created once a
    ``wx.glcanvas.GLCanvas`` has been created, and is visible on screen.  The
    ``GLContext`` class therefore creates a dummy ``wx.Frame`` and
    ``GLCanvas``, and displays it, before creating the ``wx`` GL context.


    A reference to this dummy ``wx.Frame`` is retained, because destroying it
    can result in ``GLXBadCurrentWindow`` errors when running on
    macOS+XQuartz. The frame is destroyed on calls to the ``GLContext.destroy``
    method.


    Because ``wx`` contexts may be used even when an off-screen rendering
    context has been requested, the ``GLContext`` class has the ability to
    create and run a temporary ``wx.App``, on which the canvas and context
    creation process is executed. This horrible ability is necessary, because
    a ``wx.GLContext`` must be created on a ``wx`` application loop. We cannot
    otherwise guarantee that the ``wx.GLCanvas`` will be visible before the
    ``wx.GLContext`` is created.


    The above issue has the effect that the real underlying ``wx.GLContext``
    may only be created after the ``GLContext.__init__`` method has returned.
    Therefore, you must use the ``ready`` callback function if you are
    creating a ``wx`` GL context - this function will be called when the
    ``GLContext`` is ready to be used.


    You can get away without using the ``ready`` callback in the following
    situations:

      - When you are 100% sure that you will be using OSMesa.

      - When there is not (and never will be) a ``wx.MainLoop`` running, and
        you pass in ``createApp=True``.
    """

    def __init__(self,
                 offscreen=False,
                 other=None,
                 target=None,
                 createApp=False,
                 ready=None,
                 raiseErrors=False):
        """Create a ``GLContext``.

        :arg offscreen:   On-screen or off-screen context?

        :arg other:       Another ``GLContext`` instance with which GL state
                          should be shared.

        :arg target:      If ``other`` is not ``None``, this must be a
                          reference to a ``WXGLCanvasTarget``, the rendering
                          target for the new context.

        :arg createApp:   If ``True``, and if possible, this ``GLContext`` will
                          create and run a ``wx.App`` so that it can create a
                          ``wx.glcanvas.GLContext``.

        :arg ready:       Function which will be called when the context has
                          been created and is ready to use.

        :are raiseErrors: Defaults to ``False``. If ``True``, and if the
                          ``ready`` function raises an error, that error is not
                          caught.
        """

        def defaultReady():
            pass

        if ready is None:
            ready = defaultReady

        self.__offscreen = offscreen
        self.__ownApp    = False
        self.__context   = None
        self.__canvas    = None
        self.__parent    = None
        self.__buffer    = None
        self.__app       = None

        osmesa     = os.environ.get('PYOPENGL_PLATFORM', None) == 'osmesa'
        canHaveGui = fwidgets.canHaveGui()
        haveGui    = fwidgets.haveGui()

        # On-screen contexts *must* be
        # created via a wx event loop
        if (not offscreen) and not (haveGui or createApp):
            raise ValueError('On-screen GL contexts must be '
                             'created on the wx.MainLoop')

        # For off-screen, only use OSMesa
        # if we have no cnoice, or if
        # dictated by PYOPENGL_PLATFORM.
        # Otherewise we use wx if possible.
        if offscreen and (osmesa or (not canHaveGui)):
            self.__createOSMesaContext()
            ready()
            return

        self.__ownApp = (not haveGui) and createApp

        # A context already exists - we don't
        # need to create a GL canvas to create
        # another one.
        if other is not None:
            self.__createWXGLContext(other=other.__context, target=target)
            ready()
            return

        # Create a wx.App if we've been
        # given permission to do so
        # (via the createApp argument)
        if self.__ownApp:
            log.debug('Creating temporary wx.App')

            import fsleyes.main as fm
            self.__app = fm.FSLeyesApp()

        # Create a parent for the GL
        # canvas, and the canvas itself
        self.__createWXGLParent()
        self.__createWXGLCanvas()

        # This function creates the context
        # and does some clean-up afterwards.
        # It gets scheduled on the wx idle
        # loop.
        def create():

            app = self.__app

            self.__createWXGLContext()

            # If we've created and started
            # our own loop, kill it
            if self.__ownApp:
                log.debug('Exiting temporary wx.MainLoop')
                app.ExitMainLoop()

            if ready is not None:

                try:
                    ready()

                except Exception as e:
                    log.warning('GLContext callback function raised '
                                '{}: {}'.format(type(e).__name__,
                                                str(e)),
                                                exc_info=True)
                    if raiseErrors:
                        raise e

            # We keep the parent around, because
            # destroying it can cause GLXBadCurrentWindow
            # errors when running on macOS+XQuartz. It is
            # destroyed when the GLContext.destroy() method
            # is called.
            self.__parent.Hide()

        # If we've created our own wx.App, run its
        # main loop - we need to run the loop
        # in order to display the GL canvas and
        # context. But we can kill the loop as soon
        # as this is done (in the create function
        # above).  If an existing wx.App is running,
        # we just schedule the context creation
        # routine on it.
        idle.idle(create, alwaysQueue=True)

        if self.__ownApp:
            log.debug('Starting temporary wx.MainLoop')
            self.__app.MainLoop()


    def destroy(self):
        """Called by the module-level :func:`shutdown` function. If this is an
        on-screen context, the dummy canvas and frame that were created at
        initialisation are destroyed.
        """

        # We need to destroy the OSMesa context,
        # otherwise it will stay in memory
        if self.__buffer is not None:
            import OpenGL.raw.osmesa.mesa as osmesa
            osmesa.OSMesaDestroyContext(self.__context)

        # Clear refs to wx frame/canvas
        # before destroying the wx.App,
        # as problems can otherwise occur
        app            = self.__app
        self.__app     = None
        self.__context = None
        self.__buffer  = None
        self.__parent  = None
        self.__canvas  = None

        if app is not None:
            app.Destroy()


    def setTarget(self, target=None):
        """Set the given ``WXGLCanvasTarget`` or ``OffScreenCanvasTarget`` as
        the target for GL rendering with this context.

        If ``target`` is None, and this is an on-screen rendering context,
        the dummy ``wx.glcanvas.GLCanvas`` that was used to create this
        ``GLContext`` is set as the rendering target.
        """
        # not necessary for offscreen rendering
        if self.__offscreen:
            return True

        # destroy() has been called
        if self.__context is None:
            return False

        if target is None and self.__canvas is not None:
            return self.__context.SetCurrent(self.__canvas)

        else:
            import wx.glcanvas as wxgl
            if isinstance(target, wxgl.GLCanvas):
                return self.__context.SetCurrent(target)


    def __createWXGLParent(self):
        """Create a dummy ``wx.Frame`` to be used as the parent for the
        dummy ``wx.glcanvas.GLCanvas``.
        """

        import wx

        # Override ShouldPreventAppExit, meaning
        # that the wx.App.MainLoop will exit even
        # if a DummyFrame still exists. The wx
        # equivalent of marking a thread as a
        # daemon.
        class DummyFrame(wx.Frame):
            def ShouldPreventAppExit(self):
                return False

        log.debug('Creating dummy wx.Frame for GL context creation')

        self.__parent = DummyFrame(None, style=0)
        self.__parent.SetSize((0, 0))
        self.__parent.Show(True)


    def __createWXGLCanvas(self):
        """Create a dummy ``wx.glcanvas.GLCanvas`` instance which is to
        be used to create a context. Assigns the canvas to an attributed
        called ``__canvas``.
        """

        import wx.glcanvas as wxgl

        log.debug('Creating temporary wx.GLCanvas')

        # There's something wrong with wxPython's
        # GLCanvas (on OSX at least) - the pixel
        # format attributes have to be set on the
        # *first* GLCanvas that is created -
        # setting them on subsequent canvases will
        # have no effect. But if you set them on
        # the first canvas, all canvases that are
        # subsequently created will inherit the
        # same properties.
        attrs = WXGLCanvasTarget.displayAttributes()

        # GLCanvas initialisation with an attribute
        # list fails when running in a nomachine-like
        # remote desktop session. No idea why.
        try:
            self.__canvas = wxgl.GLCanvas(self.__parent, **attrs)
            self.__canvas.SetSize((0, 0))

        # Creating without attribute list works ok
        # though. This does mean that we don't have
        # control over depth/stencil buffer sizes,
        # under these remote desktop environments.
        except Exception:
            WXGLCanvasTarget.displayAttributes.disabled = True
            self.__canvas = wxgl.GLCanvas(self.__parent)
            self.__canvas.SetSize((0, 0))

        self.__canvas.Show(True)


    def __createWXGLContext(self, other=None, target=None):
        """Creates a ``wx.glcanvas.GLContext`` object, assigning it to
        an attribute called ``__context``. Assumes that a
        ``wx.glcanvas.GLCanvas`` has already been created.

        :arg other:  Another `wx.glcanvas.GLContext`` instance with which
                     the new context should share GL state.

        :arg target: If ``other`` is not ``None``, this must be a
                     ``wx.glcanvas.GLCanvas``, the rendering target for the
                     new context.

        .. warning:: This method *must* be called via the ``wx.MainLoop``.
        """

        import                wx
        import wx.glcanvas as wxgl

        # Versions of wxwidgets 3.1 and newer (approximately
        # corresponding to wxpython 4.1 and newer) allow
        # us to select a GL compatibility profile (required,
        # because we rely on GL 1.4/2.1).
        #
        # This only seems to work with wayland/EGL builds of
        # wxPython. We assume that gtk3+wxpython>=4.1.1
        # supports requesting a compatibility profile,
        # but don't bother for gtk2+older wxpython
        wxver  = fwidgets.wxVersion()
        wxplat = fwidgets.wxPlatform()
        if wxver is not None                               and \
           fslversion.compareVersions(wxver, '4.1.1') >= 0 and \
           wxplat in (fwidgets.WX_MAC_COCOA, fwidgets.WX_GTK3):
            attrs = wxgl.GLContextAttrs()
            attrs.CompatibilityProfile()
            attrs.EndList()
            kwargs = {'ctxAttrs' : attrs}
        else:
            kwargs = {}

        log.debug('Creating wx.GLContext')

        if other is not None:
            self.__context = wxgl.GLContext(target, other=other, **kwargs)

        else:
            self.__context = wxgl.GLContext(self.__canvas, **kwargs)

            # We can't set the context target
            # until the dummy canvas is
            # physically shown on the screen.
            while not self.__canvas.IsShownOnScreen():
                wx.GetApp().Yield()

            self.__context.SetCurrent(self.__canvas)


    def __createOSMesaContext(self):
        """Creates an OSMesa context, assigning it to an attribute called
        ``__context``.
        """

        import OpenGL.GL              as gl
        import OpenGL.raw.osmesa.mesa as osmesa
        import OpenGL.arrays          as glarrays

        log.debug('Creating gl.OSMesaContext')

        # We have to create a dummy buffer
        # for the off-screen context.
        buffer  = glarrays.GLubyteArray.zeros((640, 480, 4))
        context = osmesa.OSMesaCreateContextExt(
            gl.GL_RGBA, 8, 4, 0, None)
        osmesa.OSMesaMakeCurrent(context,
                                 buffer,
                                 gl.GL_UNSIGNED_BYTE,
                                 640,
                                 480)

        self.__buffer  = buffer
        self.__context = context


class OffScreenCanvasTarget:
    """Base class for canvas objects which support off-screen rendering. """

    def __init__(self, width, height):
        """Create an ``OffScreenCanvasTarget``. A :class:`.RenderTexture` is
        created, to be used as the rendering target.

        :arg width:    Width in pixels
        :arg height:   Height in pixels
        """

        from fsleyes.gl.textures import RenderTexture

        self.__width  = width
        self.__height = height
        self.__target = RenderTexture(
            '{}({})_RenderTexture'.format(
                type(self).__name__,
                id(self)))


    def _setGLContext(self):
        """Configures the GL context to render to this canvas. """
        return getGLContext().setTarget(self)


    def _draw(self, *a):
        """Must be provided by subclasses."""
        raise NotImplementedError()


    def getAnnotations(self):
        """Must be provided by subclasses."""
        raise NotImplementedError()


    def canvasToWorld(self, xpos, ypos):
        """Convert X/Y pixel coordinates into a location in the display
        coordinate system. Must be provided by subclasses.
        """
        raise NotImplementedError()


    def worldToCanvas(self, pos):
        """Convert a location in the display coordinate system into X/Y pixel
        coordinates. Must be provided by subclasses.
        """
        raise NotImplementedError()


    def GetSize(self):
        """Returns a tuple containing the canvas width and height."""
        return self.__width, self.__height


    def GetScaledSize(self):
        """Returns a tuple containing the canvas width and height."""
        return self.GetSize()


    def Refresh(self, *a):
        """Does nothing. This canvas is for static (i.e. unchanging) rendering.
        """
        pass


    def FreezeDraw(self):
        """Does nothing. This canvas is for static (i.e. unchanging) rendering.
        """
        pass


    def ThawDraw(self):
        """Does nothing. This canvas is for static (i.e. unchanging) rendering.
        """
        pass

    def FreezeSwapBuffers(self):
        """Does nothing. This canvas is for static (i.e. unchanging) rendering.
        """
        pass


    def ThawSwapBuffers(self):
        """Does nothing. This canvas is for static (i.e. unchanging) rendering.
        """
        pass


    def EnableHighDPI(self):
        """Does nothing. This canvas is for static (i.e. unchanging) rendering.
        """
        pass


    def draw(self):
        """Calls the :meth:`_draw` method, which must be provided by
        subclasses.
        """

        self._setGLContext()
        self._initGL()
        self.__target.shape = self.__width, self.__height

        with self.__target.target():
            self._draw()


    def getBitmap(self):
        """Return a (height*width*4) shaped numpy array containing the
        rendered scene as an RGBA bitmap. The bitmap will be full of
        zeros if the scene has not been drawn (via a call to
        :meth:`draw`).
        """

        self._setGLContext()
        return self.__target.getBitmap()


    def saveToFile(self, filename):
        """Saves the contents of this canvas as an image, to the specified
        file.
        """
        import matplotlib.image as mplimg
        mplimg.imsave(filename, self.getBitmap())


class WXGLCanvasTarget:
    """Base class for :class:`wx.glcanvas.GLCanvas` objects.

    It is assumed that subclasses of this base class are also subclasses of
    :class:`wx.glcanvas.GLCanvas`. Sub-classes must override the following
    methods:

    .. autosummary::
       :nosignatures:

       _initGL
       _draw

    And must also ensure that the :meth:`destroy` method is called when
    the class is being destroyed.
    """


    @staticmethod
    def canToggleHighDPI():
        """Return ``True`` if high-DPI scaling can be toggled, ``False``
        otherwise.

        Under GTK, high DPI support is not possible with wxPython < 4.0.7, as
        ``wx.Window.GetContentScaleFactor`` always returns 1. Under GTK and
        from wxPython 4.0.7 onwards, GL canvases are scaled automatically.

        Under macOS and wxpython < 4.1.0, high DPI must be explicitly requested
        for GL canvases via a Cocoa API call. This can be done via the
        :meth:`EnableHighDPI` method.

        Under macOS and with wxpython >= 4.1.0, GL canvases are scaled
        automatically.
        """

        import wx

        if platform.system() != 'Darwin':
            return False

        wxver = getattr(wx, '__version__', '1.0.0')
        wxver = [int(v) for v in wxver.split('.')[:3]]

        return wxver < [4, 1, 0]


    @staticmethod
    def displayAttributes():
        """Used within ``__init__`` methods of ``WXGLCanvasTarget``
        sub-classes.

        Return a dict to be passed as keyword arguments to the
        ``wx.glcanvas.GLCanvas.__init__`` method, defining desired OpenGL
        display attributes (e.g. depth/stencil buffer sizes).

        The ``GLCanvas`` interface changed between wxWidgets 3.0.x and 3.1.x
        (roughly corresponding to wxPython 4.0.x and 4.1.x) - this method
        checks the wxPython version and returns a suitable set of arguments.

        In certain remote environments (e.g. x2go), this method may return
        an empty dictionary, as in some environments it does not seem to be
        possible to request display attributes.
        """

        # The GLContext.__createWXGLCanvas method
        # might set a disabled flag on this method
        # if it appears that we can't request
        # specific values for display attributes
        if getattr(WXGLCanvasTarget.displayAttributes, 'disabled', False):
            return {}

        import wx
        import wx.glcanvas as wxgl

        # the format of wx.__version__ is not
        # consistent (e.g. "4.0.7.post2", "4.1.0", etc)
        try:
            version = '.'.join(wx.__version__.split('.')[:3])
        except Exception:
            version = '4.0.0'

        # Use new API for 4.1.0 and newer
        if fslversion.compareVersions(version, '4.1.0') >= 0:
            attrs = wxgl.GLAttributes()
            attrs.MinRGBA(8, 8, 8, 8) \
                 .DoubleBuffer()      \
                 .Depth(24)           \
                 .Stencil(4)          \
                 .EndList()
            kwargs = {'dispAttrs' : attrs}

        else:
            attrs = [wxgl.WX_GL_RGBA,
                     wxgl.WX_GL_DOUBLEBUFFER,
                     wxgl.WX_GL_STENCIL_SIZE, 4,
                     wxgl.WX_GL_DEPTH_SIZE,   24,
                     0,
                     0]
            kwargs = {'attribList' : attrs}
        return kwargs


    def __init__(self):
        """Create a ``WXGLCanvasTarget``. """

        import wx

        context = getGLContext()

        # If we are on OSX, and using the Apple Software
        # Renderer (e.g. in a virtual machine), then it
        # seems that we have to have a separate context
        # for each display. If we don't, then strange
        # refresh problems will occur.
        if platform.system() == 'Darwin' and \
           'software' in GL_RENDERER.lower():

            log.debug('Creating separate GL context for '
                      'WXGLCanvasTarget {}'.format(id(self)))

            context = GLContext(other=context, target=self)

        self.__glReady           = False
        self.__freezeDraw        = False
        self.__freezeSwapBuffers = False
        self.__context           = context

        if WXGLCanvasTarget.canToggleHighDPI():
            self.__dpiscale = 1.0
        else:
            self.__dpiscale = self.GetContentScaleFactor()

        self.Bind(wx.EVT_PAINT,            self.__onPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.__onEraseBackground)


    def destroy(self):
        """Must be called when this WXGLCanvasTarget is no longer in use.
        Clears the GL rendering context target.
        """
        self.__context.setTarget(None)
        self.__context = None


    @property
    def destroyed(self):
        """Returns ``True`` if :meth:`destroy` has been called, ``False``
        otherwise.
        """
        return self.__context is None


    def __onEraseBackground(self, ev):
        """Called on ``wx.EVT_ERASE_BACKGROUND`` events. Does nothing. """
        pass


    def __onPaint(self, ev):
        """Called on ``wx.EVT_PAINT`` events. Schedules :meth:`Refresh`
        to be called on the idle loop.
        """

        def doRefresh():
            if fwidgets.isalive(self):
                self.Refresh()

        # GL canvases do need to be refreshed
        # on EVT_PAINT events. If they are not,
        # the canvas will be corrupted.
        if not self.__freezeDraw:
            idle.idle(doRefresh)


    def _initGL(self):
        """This method should perform any OpenGL data initialisation required
        for rendering. Must be implemented by subclasses.
        """
        raise NotImplementedError()


    def getAnnotations(self):
        """Must be provided by subclasses."""
        raise NotImplementedError()


    def canvasToWorld(self, xpos, ypos):
        """Convert X/Y pixel coordinates into a location in the display
        coordinate system. Must be provided by subclasses.
        """
        raise NotImplementedError()


    def worldToCanvas(self, pos):
        """Convert a location in the display coordinate system into X/Y pixel
        coordinates. Must be provided by subclasses.
        """
        raise NotImplementedError()


    def _draw(self, *a):
        """This method should implement the OpenGL drawing logic - it must be
        implemented by subclasses.

        .. note:: When runing with an on-screen display, this method should
                  never be called directly - call the :meth:`Refresh` method
                  instead.
        """
        raise NotImplementedError()


    def __realDraw(self, *a):
        """Called when the canvas needs to be refreshed.

        This method calls :meth:`_initGL` if it has not already been called.
        Otherwise, it calls the subclass :meth:`_draw` method.
        """

        # The only purpose of __realDraw is to
        # make sure that _initGL has been called
        # before the first call to _draw.  So
        # after initGL has been called, we replace
        # this method (__realDraw) and _draw
        # with drawWrapper. This wrapper calls
        # the original _draw method (provided by
        # subclasses), and manages GL flushing
        # and front/back buffer swaps. It also
        # honours FreezeDraw and FreezeSwapBuffers.
        subClassDraw = self._draw

        import OpenGL.GL as gl

        def drawWrapper(*a, **kwa):

            if self.destroyed:
                return

            if not self.__freezeDraw:
                subClassDraw(*a, **kwa)

            if not self.__freezeSwapBuffers:

                # If a draw didn't occur above,
                # this canvas may not have been
                # set as the target for the GL
                # context.
                if not self.__freezeDraw:
                    self._setGLContext()
                self.SwapBuffers()

            # If not swapping the front/back
            # buffers immediately after a draw,
            # make sure we flush all recent GL
            # operations, otherwise we may lose
            # them.
            elif not self.__freezeDraw:
                gl.glFlush()

        def doInit(*a):

            import wx

            try:
                self.__glReady = True
                self._initGL()

                self.__realDraw = drawWrapper
                self.__draw     = drawWrapper
                self._draw()

            # Just in case this canvas
            # was destroyed before it
            # had a chance to be drawn
            # on - wxpython raises a
            # RuntimeError on attempts
            # to use deleted wx objects
            except RuntimeError:
                pass

        if not self.__glReady:
            import wx
            wx.CallAfter(doInit)


    def _setGLContext(self):
        """Configures the GL context for drawing to this canvas.

        This method should be called before any OpenGL operations related to
        this canvas take place (e.g. texture/data creation, drawing, etc).
        """
        if not (fwidgets.isalive(self) and self.IsShownOnScreen()):
            return False

        log.debug('Setting context target to {} ({})'.format(
            type(self).__name__, id(self)))

        return self.__context.setTarget(self)


    def GetSize(self):
        """Returns the current canvas size. """
        return self.GetClientSize().Get()


    def GetContentScaleFactor(self):
        """Overrides ``wx.Window.GetContentScaleFactor``.

        Calls the base class implementation, except where wxpython 3.0.2.0 is
        being used, as the method does not exist in that version. In this
        case, 1.0 is returned.
        """
        try:
            return float(super().GetContentScaleFactor())
        except AttributeError:
            return 1.0


    def GetScale(self):
        """Returns the current DPI scaling factor. """

        scale = self.GetContentScaleFactor()

        # Reset scaling factor in case the canvas
        # window has moved between displays with
        # different scaling factors
        if scale == 1 and self.__dpiscale != 1:
            self.EnableHighDPI(False)

        return self.__dpiscale


    def GetScaledSize(self):
        """Returns the current canvas size, scaled by the current DPI scaling
        factor.
        """
        w, h = self.GetSize()
        s    = self.GetScale()

        return int(round(w * s)), int(round(h * s))


    def Refresh(self, *a):
        """Triggers a redraw via the :meth:`_draw` method. """
        self.__realDraw()


    def FreezeDraw(self):
        """*Freezes* updates to the canvas. See :meth:`ThawDraw`. """
        self.__freezeDraw = True


    def ThawDraw(self):
        """Unfreezes canvas updates. See :meth:`FreezeDraw`. """
        self.__freezeDraw = False


    def FreezeSwapBuffers(self):
        """*Freezes* canvas fron/back buffer swaps, but not canvas drawing.
        See :meth:`ThawSwapBuffers`.
        """
        self.__freezeSwapBuffers = True


    def ThawSwapBuffers(self):
        """Unfreezes canvas fron/back buffer swaps. See
        :meth:`FreezeSwapBuffers`.
        """
        self.__freezeSwapBuffers = False


    def SwapBuffers(self):
        """Overrides ``wx.GLCanvas.SwapBuffers``. Calls that method, but
        only if ``FreezeSwapBuffers`` is not active.
        """
        if not self.__freezeSwapBuffers:
            if self._setGLContext():
                super(WXGLCanvasTarget, self).SwapBuffers()


    def EnableHighDPI(self, enable=True):
        """Attempts to enable/disable high-resolution rendering.
        """

        # We don't necessarily need to
        # enable high-DPI support - see
        # __init__.
        if not WXGLCanvasTarget.canToggleHighDPI():
            return

        if not self._setGLContext():
            return

        self.__dpiscale = 1.0

        # If the display can't scale,
        # (scale == 1) there's no point
        # in enabling it.
        scale  = self.GetContentScaleFactor()
        enable = enable and scale > 1

        # on macOS, we have to set
        # scaling on the GL canvas
        try:
            import objc
            nsview = objc.objc_object(c_void_p=self.GetHandle())
            nsview.setWantsBestResolutionOpenGLSurface_(enable)

        # objc library not present
        except ImportError:
            return

        if enable: self.__dpiscale = scale
        else:      self.__dpiscale = 1.0


    def getBitmap(self):
        """Return a (width*height*4) shaped numpy array containing the
        rendered scene as an RGBA bitmap.
        """
        import OpenGL.GL as gl
        import numpy     as np

        self._setGLContext()

        width, height = self.GetScaledSize()

        # Make sure we're reading
        # from the front buffer
        gl.glReadBuffer(gl.GL_FRONT_LEFT)

        bmp = gl.glReadPixels(
            0, 0,
            width, height,
            gl.GL_RGBA,
            gl.GL_UNSIGNED_BYTE)

        bmp = np.frombuffer(bmp, dtype=np.uint8)
        bmp = bmp.reshape((height, width, 4))
        bmp = np.flipud(bmp)

        return bmp



#!/usr/bin/env python
#
# atlasmanagementpanel.py - The AtlasManagementPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`AtlasManagementPanel`, which is a sub-panel
that is used by the :class:`.AtlasPanel`.
"""


import logging

import wx

import fsleyes_widgets.elistbox  as elistbox
import fsl.data.atlases          as atlases
import fsleyes.panel             as fslpanel
import fsleyes.actions.loadatlas as loadatlas


log = logging.getLogger(__name__)


class AtlasManagementPanel(fslpanel.FSLeyesPanel):
    """The ``AtlasManagementPanel`` is a sub-panel used by the
    :class:`.AtlasPanel`. It simply displays a list of all known atlases,
    and allows the user to add/remove atlases to/from the list.
    """


    def __init__(self, parent, overlayList, displayCtx, atlasPanel):
        """Create an ``AtlasManagementPanel``.

        :arg parent:      the :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg atlasPanel:  The :class:`.AtlasPanel` instance that has created
                          this ``AtlasManagementPanel``.
        """
        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, atlasPanel.frame)

        descs = atlases.listAtlases()
        names = [d.name     for d in descs]
        paths = [d.specPath for d in descs]

        self.__atlasList = elistbox.EditableListBox(
            self,
            labels=names,
            clientData=descs,
            tooltips=paths,
            vgap=5,
            style=(elistbox.ELB_NO_MOVE |
                   elistbox.ELB_TOOLTIP_DOWN))

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__atlasList, flag=wx.EXPAND, proportion=1)

        self.SetSizer(self.__sizer)
        self.SetMinSize(self.__sizer.GetMinSize())

        self.__atlasList.Bind(elistbox.EVT_ELB_ADD_EVENT,
                              self.__onListAdd)
        self.__atlasList.Bind(elistbox.EVT_ELB_REMOVE_EVENT,
                              self.__onListRemove)

        atlases.registry.register(self.name, self.__atlasAdded,   'add')
        atlases.registry.register(self.name, self.__atlasRemoved, 'remove')


    def destroy(self):
        """Must be called when this ``AtlasManagementPanel`` is no longer
        needed. Removes some property/notification listeners, and calls
        the base class ``destroy`` method.
        """

        fslpanel.FSLeyesPanel.destroy(self)

        atlases.registry.deregister(self.name, 'add')
        atlases.registry.deregister(self.name, 'remove')


    def __atlasAdded(self, registry, topic, desc):
        """Called when an atlas is removed from the :class:`.AtlasRegistry`.
        Removes the corresponding atlas from the list.
        """

        atlasID  = desc.atlasID
        allIDs   = [d.atlasID for d in registry.listAtlases()]
        index    = allIDs.index(atlasID)

        self.__atlasList.Insert(desc.name,
                                index,
                                clientData=desc,
                                tooltip=desc.specPath)


    def __atlasRemoved(self, registry, topic, desc):
        """Called when an atlas is removed from the :class:`.AtlasRegistry`.
        Removes the corresponding atlas from the list.
        """

        index = self.__atlasList.IndexOf(desc)

        self.__atlasList.Remove(index)


    def __onListAdd(self, ev):
        """Called when the user clicks the *Add* button on the list.
        Calls the :func:`.loadatlas.loadAtlas` function.
        """
        loadatlas.loadAtlas(self.frame)


    def __onListRemove(self, ev):
        """Called when the user clicks the *Remove* button on the list.
        Removes the corresponding atlas from the :class:`.AtlasRegistry`.
        """

        with atlases.registry.skip(self.name, 'remove'):
            atlases.removeAtlas(ev.data.atlasID)



#!/usr/bin/env python
#
# updatecheck.py - The UpdateCheckAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.UpdateCheckAction`, which checks to see
if a new version of FSLeyes is available.
"""


import logging

import urllib.request as request
import ssl
import json

import wx
import wx.adv as wxadv

import fsl.version                  as fslversion

import fsleyes_widgets.utils.status as status
import fsleyes.version              as version
import fsleyes.strings              as strings
from . import                          base


log = logging.getLogger(__name__)


_FSLEYES_URL = 'https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FSLeyes'
"""A url to direct the user towards to download the latest version of FSLeyes.
"""


_FSLEYES_VERSION_URL = 'https://api.anaconda.org/package/conda-forge/fsleyes'
"""A url which points to a JSON file that contains information about the
FSLeyes package on conda-forge.
"""


class UpdateCheckAction(base.Action):
    """The :class:`.UpdateCheckAction` is an :class:`.Action` which checks to
    see if a new version of FSLeyes is available, and tells the user if there
    is.
    """


    def __init__(self, overlayList, displayCtx):
        """Create an ``UpdateCheckAction``. """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__checkForUpdates)


    def __checkForUpdates(self,
                          showUpToDateMessage=True,
                          showErrorMessage=True,
                          ignorePoint=False):
        """Run this action. Downloads a text file from a URL which contains
        the latest available version of FSLeyes. Compares that version with
        the running version. Displays a message to the user.

        :arg showUpToDateMessage: Defaults to ``True``. If ``False``, and
                                  the current version of FSLeyes is up to
                                  date, the user is not informed.

        :arg showErrorMessage:    Defaults to ``True``. If ``False``, and
                                  some error occurs while checking for
                                  updates, the user is not informed.

        :arg ignorePoint:         Defaults to ``False``. If ``True``, the
                                  point release number is ignored in the
                                  comparison.
        """

        errMsg   = strings.messages[self, 'newVersionError']
        errTitle = strings.titles[  self, 'newVersionError']

        with status.reportIfError(errTitle,
                                  errMsg,
                                  raiseError=False,
                                  report=showErrorMessage):

            log.debug('Checking for FSLeyes updates ({})'.format(
                _FSLEYES_VERSION_URL))

            ctx                = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode    = ssl.CERT_NONE

            with request.urlopen(_FSLEYES_VERSION_URL, context=ctx) as f:
                info   = json.loads(f.read().decode('utf-8'))
                latest = info['latest_version']

            current  = version.__version__
            upToDate = fslversion.compareVersions(latest,
                                                  current,
                                                  ignorePoint) <= 0

            log.debug('This version of FSLeyes ({}) is '
                      '{} date (latest: {})'.format(
                          current,
                          'up to' if upToDate else 'out of',
                          latest))

            if upToDate and not showUpToDateMessage:
                return

            urlMsg = strings.messages[self, 'updateUrl']

            if upToDate:
                title = strings.titles[  self, 'upToDate']
                msg   = strings.messages[self, 'upToDate']
                msg   = msg.format(current)

            else:
                title = strings.titles[  self, 'newVersionAvailable']
                msg   = strings.messages[self, 'newVersionAvailable']
                msg   = msg.format(current, latest, _FSLEYES_URL)

            parent = wx.GetTopLevelWindows()[0]
            dlg    = UrlDialog(parent, title, msg, urlMsg, _FSLEYES_URL)

            dlg.CentreOnParent()
            dlg.ShowModal()


class UrlDialog(wx.Dialog):
    """Custom ``wx.Dialog`` used by the :class:`UpdateCheckAction` to
    display a message containing the FSLeyes download URL to the user.
    """


    def __init__(self,
                 parent,
                 title,
                 msg,
                 urlMsg=None,
                 url=None):
        """Create a ``UrlDialog``.

        :arg parent: ``wx`` parent object
        :arg title:  Dialog title
        :arg msg:    Message to display
        :arg urlMsg: Message to display next to the URL. Not shown if a URL
                     is not provided.
        :arg url:    URL to display.
        """

        wx.Dialog.__init__(self,
                           parent,
                           title=title,
                           style=wx.DEFAULT_DIALOG_STYLE)

        ok  = wx.Button(    self, label='Ok', id=wx.ID_OK)
        msg = wx.StaticText(self, label=msg)

        self.__ok = ok

        if urlMsg is not None:
            urlMsg = wx.StaticText(self, label=urlMsg)
        if url is not None:
            url = wxadv.HyperlinkCtrl(self, url=url)

        sizer    = wx.BoxSizer(wx.VERTICAL)
        btnSizer = wx.BoxSizer(wx.HORIZONTAL)

        sizer.Add((1, 20),  flag=wx.EXPAND, proportion=1)
        sizer.Add(msg,      flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=20)
        sizer.Add((1, 20),  flag=wx.EXPAND, proportion=1)

        if urlMsg is not None and url is not None:
            sizer.Add(urlMsg,  flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=20)
            sizer.Add((1, 5),  flag=wx.EXPAND, proportion=1)

        if url is not None:
            sizer.Add(url,     flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=20)
            sizer.Add((1, 20), flag=wx.EXPAND, proportion=1)

        btnSizer.Add((20, 1),  flag=wx.EXPAND, proportion=1)
        btnSizer.Add(ok,       flag=wx.EXPAND)
        btnSizer.Add((20, 1),  flag=wx.EXPAND)

        sizer.Add(btnSizer, flag=wx.EXPAND)
        sizer.Add((1, 20),  flag=wx.EXPAND, proportion=1)

        ok.SetDefault()
        self.SetSizer(sizer)
        self.Layout()
        self.Fit()


    @property
    def ok(self):
        """Return a reference to the OK button. """
        return self.__ok



#!/usr/bin/env python
#
# displaycontext.py - The DisplayContext class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`DisplayContext` class, which defines
general display settings for displaying the overlays in a
:class:`.OverlayList`.
"""


import sys
import logging
import contextlib

import numpy        as np
import numpy.linalg as npla

import        fsl.data.image as fslimage
import        fsleyes_props  as props
from . import group          as dcgroup


log = logging.getLogger(__name__)


class InvalidOverlayError(Exception):
    """An error raised by the :meth:`DisplayContext.getDisplay`
    and :meth:`DisplayContext.getOpts` methods to indicate that
    the specified overlay is not in the :class:`.OverlayList`.
    """
    pass


class DisplayContext(props.SyncableHasProperties):
    """A ``DisplayContext`` instance contains a number of properties defining
    how the overlays in an :class:`.OverlayList` are to be displayed, and
    related contextual information.


    A ``DisplayContext`` instance is responsible for creating and destroying
    :class:`.Display` instances for every overlay in the
    ``OverlayList``. These ``Display`` instances, and the corresponding
    :class:`.DisplayOpts` instances (which, in turn, are created/destroyed by
    ``Display`` instances) can be accessed with the :meth:`getDisplay` and
    :meth:`getOpts` method respectively.


    A number of other useful methods are provided by a ``DisplayContext``
    instance:

    .. autosummary::
       :nosignatures:

        getDisplay
        getOpts
        getReferenceImage
        displayToWorld
        worldToDisplay
        displaySpaceIsRadiological
        selectOverlay
        getSelectedOverlay
        getOverlayOrder
        getOrderedOverlays
        freeze
        freezeOverlay
        thawOverlay
        defaultDisplaySpace
        detachDisplaySpace
    """


    selectedOverlay = props.Int(minval=0, default=0, clamped=True)
    """Index of the currently 'selected' overlay.

    .. note:: The value of this index is in relation to the
              :class:`.OverlayList`, rather than to the :attr:`overlayOrder`
              list.

              If you're interested in the currently selected overlay, you must
              also listen for changes to the :attr:`.OverlayList.images` list
              as, if the list changes, the :attr:`selectedOverlay` index may
              not change, but the overlay to which it points may be different.
    """


    location = props.Point(ndims=3)
    """The location property contains the currently selected 3D location (xyz)
    in the display coordinate system. Different ``DisplayContext`` instances
    may be using different display coordinate systems - see the
    :attr:`displaySpace` property.
    """


    worldLocation = props.Point(ndims=3)
    """The :attr:`location` property contains the currently selected 3D
    location (xyz) in the current display coordinate system. Whenever the
    :attr:`location` changes, it gets transformed into the world coordinate
    system, and propagated to this property. The location of different
    ``DisplayContext`` instances is synchronised through this property.

    .. note:: If any :attr:`.NiftiOpts.transform` properties have been modified
              independently of the :attr:`displaySpace`, this value will be
              invalid.
    """


    bounds = props.Bounds(ndims=3)
    """This property contains the min/max values of a bounding box (in display
    coordinates) which is big enough to contain all of the overlays in the
    :class:`.OverlayList`.

    .. warning:: This property shouid be treated as read-only.
    """


    overlayOrder = props.List(props.Int())
    """A list of indices into the :attr:`.OverlayList.overlays`
    list, defining the order in which the overlays are to be displayed.

    See the :meth:`getOrderedOverlays` method.
    """


    overlayGroups = props.List()
    """A list of :class:`.OverlayGroup` instances, each of which defines
    a group of overlays which share display properties.

    By default there is one overlay group, to which all new overlays are
    initially added. The :class:`.OverlayListPanel` allows the user to
    add/remove overlays from this default group.
    """


    syncOverlayDisplay = props.Boolean(default=True)
    """If this ``DisplayContext`` instance has a parent (see
    :mod:`props.syncable`), and this property is ``True``, the properties of
    the :class:`.Display` and :class:`.DisplayOpts` instances for every
    overlay managed by this ``DisplayContext`` instance will be synchronised
    to those of the parent instance. Otherwise, the display properties for
    every overlay will be unsynchronised from the parent.

    Properties which control the current volume/timepoint in a 4D data set are
    not managed by this property - see the :attr: :attr:`syncOverlayVolume`
    property.

    Synchronisation of the following properties between child and parent
    ``DisplayContext`` instances is also controlled by this flag:

      - :attr:`displaySpace`
      - :attr:`bounds`
      - :attr:`radioOrientation`

    .. note:: This property is accessed by the :class:`.Display` class, in its
              constructor, and when it creates new :class:`.DisplayOpts`
              instances, to set initial sync states.
    """


    syncOverlayVolume = props.Boolean(default=True)
    """This property performs the same task as the :attr:`syncOverlayDisplay`
    property, but it only affects :class:`DisplayOpts` properties which control
    the current volume/timepoint in a 4D overlay.
    """


    displaySpace = props.Choice(('world', 'scaledVoxel'))
    """The *space* in which overlays are displayed. This property defines the
    display coordinate system for this ``DisplayContext``. When it is changed,
    the :attr:`.NiftiOpts.transform` property of all :class:`.Nifti` overlays
    in the :class:`.OverlayList` is updated. It has two settings, described
    below. The options for this property are dynamically added by
    :meth:`__updateDisplaySpaceOptions`.

    1. **World** space (a.k.a. ``'world'``)

       All :class:`.Nifti` overlays are displayed in the space defined by
       their affine transformation matrix - the :attr:`.NiftiOpts.transform`
       property for every ``Nifti`` overlay is set to ``affine``.

    2. **Scaled-voxel** space (a.k.a. ``'scaledVoxel'``)

       All :class:`.Nifti` overlays are displayed in a scaled-voxel coordinate
       system, with origin set to the centre of voxel ``(0, 0, 0)``, and
       voxels scaled by image pixdims. This is accomplished by setting the
       :attr:`.NiftiOpts.transform` property for every ``Nifti`` overlay to
       ``pixdim``.

    3. **Reference image** space

       A single :class:`.Nifti` overlay is selected as a *reference* image,
       and is displayed in scaled voxel space (with a potential L/R flip for
       neurological images - its :attr:`.NiftiOpts.transform` is set to
       ``pixdim-flip``). All other ``Nifti`` overlays are transformed into
       this reference space - their :attr:`.NiftiOpts.transform` property is
       set to ``reference``, which results in them being transformed into the
       scaled voxel space of the reference image.

    .. note:: The :attr:`.NiftiOpts.transform` property of any
              :class:`.Nifti` overlay can be set independently of this
              property. However, whenever *this* property changes, it will
              change the ``transform`` property for every ``Nifti``, in the
              manner described above.

    The :meth:`defaultDisplaySpace` can be used to control how the
    ``displaySpace`` is initialised.
    """


    radioOrientation = props.Boolean(default=True)
    """If ``True``, 2D views will show images in radiological convention
    (i.e.subject left on the right of the display). Otherwise, they will be
    shown in neurological convention (subject left on the left).

    .. note:: This setting is not enforced by the ``DisplayContext``. It is
              the responsibility of the :class:`.OrthoPanel` and
              :class:`.LightBoxPanel` (and other potential future 2D view
              panels) to implement the flip.
    """


    autoDisplay = props.Boolean(default=False)
    """If ``True``, whenever an overlay is added to the :class:`.OverlayList`,
    the :mod:`autodisplay` module will be used to automatically configure
    its display settings. Note that the ``DisplayContext`` does not perform
    this configuration - this flag is used by other modules (e.g. the
    :class:`.OverlayListPanel` and the :class:`.OpenFileAction`).
    """


    loadInMemory = props.Boolean(default=False)
    """If ``True``, all :class:`.Image` instances will be loaded into memory,
    regardless of their size. Otherwise (the default), large compressed
    ``Image`` overlays may be kept on disk.


    .. note:: Changing the value of this property will not affect existing
              ``Image`` overlays.


    .. note:: This property may end up being used in a more general sense by
              any code which needs to decide whether to do things in a more
              or less memory-intensive manner.
    """


    def __init__(self, overlayList, parent=None, defaultDs='ref', **kwargs):
        """Create a ``DisplayContext``.

        :arg overlayList: An :class:`.OverlayList` instance.

        :arg parent:      Another ``DisplayContext`` instance to be used
                          as the parent of this instance, passed to the
                          :class:`.SyncableHasProperties` constructor.

        :arg defaultDs:   Initial value for the :meth:`defaultDisplaySpace`.
                          Either ``'ref'`` or ``'world'``. If ``'ref'`` (the
                          default), when overlays are added to an empty list,
                          the :attr:`displaySpace` will be set to the first
                          :class:`.Nifti` overlay. Otherwise (``'world'``),
                          the display space will be set to ``'world'``.

        All other arguments are passed through to the ``SyncableHasProperties``
        constructor, in addition to the following:

          - The ``syncOverlayDisplay``,, ``syncOverlayVolume``, ``location``,
            and ``bounds`` properties are added to the ``nobind`` argument

          - The ``overlayGroups``, ``autoDisplay`` and ``loadInMemory``
            properties are added to the ``nounbind`` argument.
        """

        kwargs = dict(kwargs)

        nobind   = kwargs.pop('nobind',   [])
        nounbind = kwargs.pop('nounbind', [])

        nobind  .extend(['syncOverlayDisplay',
                         'syncOverlayVolume',
                         'location',
                         'bounds'])
        nounbind.extend(['overlayGroups',
                         'autoDisplay',
                         'loadInMemory'])

        kwargs['parent']   = parent
        kwargs['nobind']   = nobind
        kwargs['nounbind'] = nounbind
        kwargs['state']    = {'overlayOrder' : False}

        props.SyncableHasProperties.__init__(self, **kwargs)

        self.__overlayList = overlayList
        self.__name        = '{}_{}'.format(self.__class__.__name__, id(self))
        self.__child       = parent is not None

        # When the first overlay(s) is/are
        # added, the display space may get
        # set either to a reference image,
        # or to world. The defaultDisplaySpace
        # controls this behaviour.
        self.defaultDisplaySpace = defaultDs

        # The overlayOrder is unsynced by
        # default, but we will inherit the
        # current parent value.
        if self.__child: self.overlayOrder[:] = parent.overlayOrder[:]
        else:            self.overlayOrder[:] = range(len(overlayList))

        # If this is the first child DC, we
        # need to initialise the display
        # space and location. If there is
        # already a child DC, then we have
        # (probably) inherited initial
        # settings.
        if self.__child:
            self.__initDS = (len(parent.getChildren()) - 1) == 0


        # While the DisplayContext may refer to
        # multiple overlay groups, we are currently
        # using just one, allowing the user to specify
        # a set of overlays for which their display
        # properties are 'locked'.
        if not self.__child:
            lockGroup = dcgroup.OverlayGroup(self, overlayList)
            self.overlayGroups.append(lockGroup)

        # This dict contains the Display
        # objects for every overlay in
        # the overlay list, as
        # {Overlay : Display} mappings
        self.__displays = {}

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged,
                                immediate=True)

        if self.__child:
            self.addListener('syncOverlayDisplay',
                             self.__name,
                             self.__syncOverlayDisplayChanged)
            self.addListener('syncOverlayVolume',
                             self.__name,
                             self.__syncOverlayVolumeChanged)
            self.addListener('displaySpace',
                             self.__name,
                             self.__displaySpaceChanged,
                             immediate=True)
            self.addListener('location',
                             self.__name,
                             self.__locationChanged,
                             immediate=True)
            self.addListener('worldLocation',
                             self.__name,
                             self.__worldLocationChanged,
                             immediate=True)

        # The overlayListChanged method
        # is important - check it out
        self.__overlayListChanged()

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message."""

        # The log atribute, and internals of the
        # logging module, be GC'd before this
        # method is called, so absorb any errors.
        try:
            log.debug('%s.del (%s)', type(self).__name__, id(self))
        except Exception:
            pass


    def destroy(self):
        """This method must be called when this ``DisplayContext`` is no
        longer needed.

        When a ``DisplayContext`` is destroyed, all of the :class:`.Display`
        instances managed by it are destroyed as well.
        """

        self.detachAllFromParent()

        overlayList = self.__overlayList
        displays    = self.__displays

        self.__overlayList = None
        self.__displays    = None

        overlayList.removeListener('overlays', self.__name)

        if self.__child:
            self.removeListener('syncOverlayDisplay', self.__name)
            self.removeListener('syncOverlayVolume',  self.__name)
            self.removeListener('displaySpace',       self.__name)
            self.removeListener('location',           self.__name)
            self.removeListener('worldLocation',      self.__name)
        else:
            for g in list(self.overlayGroups):
                self.overlayGroups.remove(g)
                g.destroy()

        for overlay, display in displays.items():
            display.destroy()


    @property
    def destroyed(self):
        """Returns ``True`` if this ``DisplayContext`` has been, or is being,
        destroyed, ``False`` otherwise.
        """
        return self.__overlayList is None


    def getDisplay(self, overlay, **kwargs):
        """Returns the :class:`.Display` instance for the specified overlay
        (or overlay index).

        If the overlay is not in the ``OverlayList``, an
        :exc:`InvalidOverlayError` is raised.  Otheriwse, if a
        :class:`Display` object does not exist for the given overlay, one is
        created.

        If this ``DisplayContext`` has been destroyed, a ``ValueError`` is
        raised.

        :arg overlay:     The overlay to retrieve a ``Display`` instance for,
                          or an index into the ``OverlayList``.

        All other keyword arguments are assumed to be ``name=value`` pairs,
        containing initial property values.
        """

        if overlay is None:
            raise ValueError('No overlay specified')

        if self.destroyed:
            raise ValueError('DisplayContext has been destroyed')

        if overlay not in self.__overlayList:
            raise InvalidOverlayError('Overlay {} is not in '
                                      'list'.format(overlay.name))

        if isinstance(overlay, int):
            overlay = self.__overlayList[overlay]

        try:
            display = self.__displays[overlay]

        except KeyError:

            if not self.__child:
                dParent = None
            else:
                dParent = self.getParent().getDisplay(overlay, **kwargs)

            from .display import Display

            display = Display(overlay,
                              self.__overlayList,
                              self,
                              parent=dParent,
                              **kwargs)
            self.__displays[overlay] = display

        return display


    def getOpts(self, overlay):
        """Returns the :class:`.DisplayOpts` instance associated with the
        specified overlay.  See :meth:`getDisplay` and :meth:`.Display.opts`
        for more details.
        """

        if overlay is None:
            raise ValueError('No overlay specified')

        if self.destroyed:
            raise ValueError('DisplayContext has been destroyed')

        if overlay not in self.__overlayList:
            raise InvalidOverlayError('Overlay {} is not in '
                                      'list'.format(overlay.name))

        return self.getDisplay(overlay).opts


    def getReferenceImage(self, overlay):
        """Convenience method which returns the reference image associated
        with the given overlay, or ``None`` if there is no reference image.

        See the :class:`.DisplayOpts.referenceImage` method.
        """
        if overlay is None:
            return None

        return self.getOpts(overlay).referenceImage


    def displaySpaceIsRadiological(self):
        """Returns ``True`` if the current :attr:`displaySpace` aligns with
        a radiological orientation. A radiological orientation is one in
        which anatomical right is shown on the left of the screen, i.e.:

          - The X axis corresponds to right -> left
          - The Y axis corresponds to posterior -> anterior
          - The Z axis corresponds to inferior -> superior
        """

        if len(self.__overlayList) == 0:
            return True

        opts  = None
        space = self.displaySpace

        # Display space is either 'world', or an image.
        # We assume that 'world' is an RAS coordinate
        # system which, if transferred directly to a
        # display coordinate system, would result in a
        # neurological view (left on left, right on
        # right).
        if space == 'world':
            return False

        # If space == scaledVoxel, we decide based on
        # the currently selected overlay.  Otherwise
        # (reference image), we decide based on the ref
        # image.
        elif space == 'scaledVoxel':
            space    = self.getSelectedOverlay()
            srcSpace = 'pixdim'
        else:
            srcSpace = 'pixdim-flip'

        # Use the FSL / FLIRT convention - if the affine
        # determinant is negative, assume neurological
        # storage order.
        if space is not None:
            ref = self.getOpts(space).referenceImage
            if ref is not None:
                opts  = self.getOpts(space)
                xform = opts.getTransform(srcSpace, 'display')
                return npla.det(xform) > 0

        # no nifti overlays loaded
        return False


    def selectOverlay(self, overlay):
        """Selects the specified ``overlay``. Raises an :exc:`IndexError` if
        the overlay is not in the list.

        If you want to select an overlay by its index in the ``OverlayList``,
        you can just assign to the :attr:`selectedOverlay` property directly.
        """
        self.selectedOverlay = self.__overlayList.index(overlay)


    def getSelectedOverlay(self):
        """Returns the currently selected overlay object,
        or ``None`` if there are no overlays.
        """
        if len(self.__overlayList) == 0:                    return None
        if self.selectedOverlay >= len(self.__overlayList): return None

        return self.__overlayList[self.selectedOverlay]


    def getOverlayOrder(self, overlay):
        """Returns the order in which the given overlay (or an index into
        the :class:`.OverlayList` list) should be displayed
        (see the :attr:`overlayOrder` property).

        Raises an :exc:`IndexError` if the overlay is not in the list.
        """
        self.__syncOverlayOrder()

        if not isinstance(overlay, int):
            overlay = self.__overlayList.index(overlay)

        return self.overlayOrder.index(overlay)


    def getOrderedOverlays(self):
        """Returns a list of overlay objects from the :class:`.OverlayList`
        list, sorted into the order that they should be displayed, as defined
        by the :attr:`overlayOrder` property.
        """
        self.__syncOverlayOrder()

        return [self.__overlayList[idx] for idx in self.overlayOrder]


    @contextlib.contextmanager
    def freeze(self, overlay):
        """This method can be used as a context manager to suppress
        notification for all :class:`.Display` and :class:`.DisplayOpts`
        properties related to the given ``overlay``::

            with displayCtx.freeze(overlay):
                # Do stuff which might trigger unwanted
                # Display/DisplayOpts notifications

        See :meth:`freezeOverlay` and :meth:`thawOverlay`.
        """
        self.freezeOverlay(overlay)

        try:
            yield

        finally:
            self.thawOverlay(overlay)


    def freezeOverlay(self, overlay):
        """Suppresses notification for all :class:`.Display` and
        :class:`.DisplayOpts` properties associated with the given ``overlay``.
        Call :meth:`.thawOverlay` to re-enable notification.

        See also the :meth:`freeze` method, which can be used as a context
        manager to automatically call this method and ``thawOverlay``.
        """
        if self.__child:
            self.getParent().freezeOverlay(overlay)
            return

        dctxs = [self] + self.getChildren()

        for dctx in dctxs:
            display = dctx.getDisplay(overlay)
            opts    = display.opts

            display.disableAllNotification()
            opts   .disableAllNotification()


    def thawOverlay(self, overlay):
        """Enables notification for all :class:`.Display` and
        :class:`.DisplayOpts` properties associated with the given ``overlay``.
        """

        if self.__child:
            self.getParent().thawOverlay(overlay)
            return
        dctxs = [self] + self.getChildren()

        for dctx in dctxs:
            display = dctx.getDisplay(overlay)
            opts    = display.opts

            display.enableAllNotification()
            opts   .enableAllNotification()


    @property
    def defaultDisplaySpace(self):
        """This property controls how the :attr:`displaySpace` is initialised
        when overlays are added to a previously empty :class:`.OverlayList`.
        If the ``defaultDisplaySpace`` is set to ``'ref'``, the
        ``displaySpace`` will be initialised to the first :class:`.Nifti`
        overlay. Otherwise (the ``defaultDisplaySpace`` is set to ``'world'``),
        the ``displaySpace`` will be set to ``'world'``.
        """
        return self.__defaultDisplaySpace


    @defaultDisplaySpace.setter
    def defaultDisplaySpace(self, ds):
        """Sets the :meth:`defaultDisplaySpace`.

        :arg ds: Either ``'ref'``, ``'scaledVoxel'``, or ``'world'``.
        """
        if ds not in ('world', 'scaledVoxel', 'ref'):
            raise ValueError('Invalid default display space: {}'.format(ds))
        self.__defaultDisplaySpace = ds


    def detachDisplaySpace(self):
        """Detaches the :attr:`displaySpace` and :attr:`bounds` properties,
        and all related :class:`.DisplayOpts` properties, from the parent
        ``DisplayContext``.

        This allows this ``DisplayContext`` to use a display coordinate
        system that is completely independent from other instances, and is not
        affected by changes to the parent properties.

        This is an irreversible operation.
        """

        self.detachFromParent('displaySpace')
        self.detachFromParent('bounds')

        for ovl in self.__overlayList:

            opts = self.getOpts(ovl)

            opts.detachFromParent('bounds')

            if isinstance(ovl, fslimage.Nifti):
                opts.detachFromParent('transform')


    def __overlayListChanged(self, *a):
        """Called when the :attr:`.OverlayList.overlays` property
        changes.

        Ensures that a :class:`.Display` and :class:`.DisplayOpts` object
        exists for every image, updates the :attr:`bounds` property, makes
        sure that the :attr:`overlayOrder` property is consistent, and updates
        constraints on the :attr:`selectedOverlay` property.
        """

        # Discard all Display instances
        # which refer to overlays that
        # are no longer in the list
        for overlay in list(self.__displays.keys()):
            if overlay not in self.__overlayList:

                display = self.__displays.pop(overlay)
                opts    = display.opts

                display.removeListener('overlayType', self.__name)
                opts   .removeListener('bounds',      self.__name)

                # The display instance will destroy the
                # opts instance, so we don't do it here
                display.destroy()

        # Ensure that a Display object exists
        # for every overlay in the list
        for overlay in self.__overlayList:

            initProps = self.__overlayList.initProps(overlay)

            # The getDisplay method
            # will create a Display object
            # if one does not already exist
            new     = overlay not in self.__displays
            display = self.getDisplay(overlay, **initProps)
            opts    = display.opts

            # Register a listener on the overlay type,
            # because when it changes, the DisplayOpts
            # instance will change, and we will need
            # to re-register the DisplayOpts.bounds
            # listener (see the next statement)
            display.addListener('overlayType',
                                self.__name,
                                self.__overlayListChanged,
                                overwrite=True)

            # Register a listener on the DisplayOpts.bounds
            # property for every overlay - if the display
            # bounds for any overlay changes, we need to
            # update our own bounds property. This is only
            # done on child DCs, as the parent DC bounds
            # only gets used for synchronisation
            if self.__child:
                opts.addListener('bounds',
                                 self.__name,
                                 self.__overlayBoundsChanged,
                                 overwrite=True)

                # If detachDisplaySpace has been called,
                # make sure the opts bounds (and related)
                # properties are also detached
                if not self.canBeSyncedToParent('displaySpace'):
                    opts.detachFromParent('bounds')
                    if isinstance(overlay, fslimage.Nifti):
                        opts.detachFromParent('transform')

            # All new overlays are initially
            # added to the default overlay group.
            # This only needs to be done on the
            # master DC, as the overlayGroups are
            # the same across all DCs.
            if new and not self.__child and len(self.overlayGroups) > 0:
                self.overlayGroups[0].addOverlay(overlay)

        # Ensure that the displaySpace
        # property options are in sync
        # with the overlay list.
        self.__updateDisplaySpaceOptions()

        # The rest of the stuff only
        # needs to be done on child DCs
        if not self.__child:
            return

        # Limit the selectedOverlay property
        # so it cannot take a value greater
        # than len(overlayList)-1. selectedOverlay
        # is always synchronised, so we only
        # need to do this on the parent DC.
        nOverlays = len(self.__overlayList)
        if nOverlays > 0:
            self.setAttribute('selectedOverlay',
                              'maxval',
                              nOverlays - 1)
        else:
            self.setAttribute('selectedOverlay', 'maxval', 0)

        # Ensure that the overlayOrder
        # property is valid
        self.__syncOverlayOrder()

        # If the overlay list was empty,
        # and is now non-empty, we need
        # to initialise the display space
        # and the display location
        initDS        = self.__initDS                      and \
                        np.all(np.isclose(self.bounds, 0)) and \
                        len(self.__overlayList) > 0
        self.__initDS = len(self.__overlayList) == 0

        # Initialise the display space. We
        # have to do this before updating
        # image transforms, and updating
        # the display bounds
        if initDS:

            displaySpace = 'world'

            if self.defaultDisplaySpace == 'ref':
                for overlay in self.__overlayList:
                    if isinstance(overlay, fslimage.Nifti):
                        displaySpace = overlay
                        break

            with props.skip(self, 'displaySpace', self.__name):
                self.displaySpace = displaySpace

        # Initialise the transform property
        # of any Image overlays which have
        # just been added to the list,
        oldList = self.__overlayList.getLastValue('overlays')[:]
        for overlay in self.__overlayList:
            if isinstance(overlay, fslimage.Nifti) and \
               (overlay not in oldList):
                self.__setTransform(overlay)

        # Ensure that the bounds
        # property is accurate
        self.__updateBounds()

        # Initialise the display location to
        # the centre of the display bounds
        if initDS:
            b = self.bounds
            self.location.xyz = [
                b.xlo + b.xlen / 2.0,
                b.ylo + b.ylen / 2.0,
                b.zlo + b.zlen / 2.0]
            self.__propagateLocation('world')
        else:
            self.__propagateLocation('display')


    def __updateDisplaySpaceOptions(self):
        """Updates the :attr:`displaySpace` property so it is synchronised with
        the current contents of the :class:`.OverlayList`

        This method is called by the :meth:`__overlayListChanged` method.
        """

        choiceProp = self.getProp('displaySpace')
        choices    = []

        for overlay in self.__overlayList:
            if isinstance(overlay, fslimage.Nifti):
                choices.append(overlay)

        choices.extend(('world', 'scaledVoxel'))

        choiceProp.setChoices(choices, instance=self)


    def __setTransform(self, image):
        """Sets the :attr:`.NiftiOpts.transform` property associated with
        the given :class:`.Nifti` overlay to a sensible value, given the
        current value of the :attr:`.displaySpace` property.

        Called by the :meth:`__displaySpaceChanged` method, and by
        :meth:`__overlayListChanged` for any :class:`.Image` overlays which
        have been newly added to the :class:`.OverlayList`.

        :arg image: An :class:`.Image` overlay.
        """

        space = self.displaySpace
        opts  = self.getOpts(image)

        # Disable notification of the bounds
        # property so the __overlayBoundsChanged
        # method does not get called. Use
        # ignoreInvalid, because this method might
        # get called before we have registered a
        # listener on the bounds property.
        with props.skip(opts, 'bounds', self.__name, ignoreInvalid=True):
            if   space == 'world':       opts.transform = 'affine'
            elif space == 'scaledVoxel': opts.transform = 'pixdim'
            elif image is space:         opts.transform = 'pixdim-flip'
            else:                        opts.transform = 'reference'


    def __displaySpaceChanged(self, *a):
        """Called when the :attr:`displaySpace` property changes. Updates the
        :attr:`.NiftiOpts.transform` property for all :class:`.Nifti`
        overlays in the :class:`.OverlayList`.
        """

        selectedOverlay = self.getSelectedOverlay()

        if selectedOverlay is None:
            return

        # Update the transform property of all
        # Image overlays to put them into the
        # new display space
        for overlay in self.__overlayList:

            if not isinstance(overlay, fslimage.Nifti):
                continue

            self.__setTransform(overlay)

        # Update the display world bounds,
        # and then update the location
        self.__updateBounds()

        # Make sure that the location is
        # kept in the same place, relative
        # to the world coordinate system
        self.__propagateLocation('display')


    def __syncOverlayOrder(self):
        """Ensures that the :attr:`overlayOrder` property is up to date
        with respect to the :class:`.OverlayList`.
        """

        if len(self.overlayOrder) == len(self.__overlayList):
            return

        #
        # NOTE: The following logic assumes that operations
        #       which modify the overlay list will only do
        #       one of the following:
        #
        #        - Adding one or more overlays to the end of the list
        #        - Removing one or more overlays from the list
        #
        # More complex overlay list modifications
        # will cause this code to break.

        oldList  = self.__overlayList.getLastValue('overlays')[:]
        oldOrder = self.overlayOrder[:]

        # If the overlay order was just the
        # list order, preserve that ordering
        if self.overlayOrder[:] == list(range(len(oldList))):
            self.overlayOrder[:] = list(range(len(self.__overlayList)))

        # If overlays have been added to
        # the overlay list, add indices
        # for them to the overlayOrder list
        elif len(oldList) < len(self.__overlayList):

            newOrder      = []
            newOverlayIdx = len(oldList)

            # The order of existing overlays is preserved,
            # and all new overlays added to the end of the
            # overlay order.
            for overlay in self.__overlayList:

                if overlay in oldList:
                    newOrder.append(oldOrder[oldList.index(overlay)])
                else:
                    newOrder.append(newOverlayIdx)
                    newOverlayIdx += 1

            self.overlayOrder[:] = newOrder

        # Otherwise, if overlays have been
        # removed from the overlay list ...
        elif len(oldList) > len(self.__overlayList):

            # Remove the corresponding indices
            # from the overlayOrder list
            for i, overlay in enumerate(oldList):
                if overlay not in self.__overlayList:
                    oldOrder.remove(i)

            # Re-generate new indices,
            # preserving the order of
            # the remaining overlays
            newOrder = [sorted(oldOrder).index(idx) for idx in oldOrder]
            self.overlayOrder[:] = newOrder


    def __overlayBoundsChanged(self, value, valid, opts, name):
        """Called when the :attr:`.DisplayOpts.bounds` property of any
        overlay changes. Updates the :attr:`bounds` property and preserves
        the display :attr:`location` in terms of the :attr:`worldLocation`.
        """

        # This method might get called
        # after DisplayOpts instance
        # has been destroyed
        if opts.display is None:
            return

        # Update the display context bounds
        # to take into account any changes
        # to individual overlay bounds.
        # Inhibit notification on the location
        # property - it will be updated properly
        # below
        self.__updateBounds()

        # Make sure the display location
        # is consistent w.r.t. the world
        # coordinate location
        self.__propagateLocation('display')


    def __syncOverlayDisplayChanged(self, *a):
        """Called when the :attr:`syncOverlayDisplay` property
        changes.

        Synchronises or unsychronises the :class:`.Display` and
        :class:`.DisplayOpts` instances for every overlay to/from their
        parent instances.
        """

        dcProps = ['displaySpace', 'bounds', 'radioOrientation']

        if self.syncOverlayDisplay:
            for p in dcProps:
                if self.canBeSyncedToParent(p):
                    self.syncToParent(p)

        else:
            for p in dcProps:
                if self.canBeUnsyncedFromParent(p):
                    self.unsyncFromParent(p)

        for display in self.__displays.values():

            if self.syncOverlayDisplay:
                display.syncAllToParent()
            else:
                display.unsyncAllFromParent()

            opts     = display.opts
            optProps = opts.getAllProperties()[0]
            exclude  = opts.getVolumeProps()

            for optProp in optProps:

                # volume properties are managed
                # by syncOverlayVolume
                if optProp in exclude:
                    continue

                if self.syncOverlayDisplay:
                    if opts.canBeSyncedToParent(optProp):
                        opts.syncToParent(optProp)
                else:
                    if opts.canBeUnsyncedFromParent(optProp):
                        opts.unsyncFromParent(optProp)


    def __syncOverlayVolumeChanged(self, *a):
        """Called when the :attr:`syncOverlayVolume` property changes.

        Synchronises or unsychronises the volume/timepoint properties of
        the :class:`.Display` and :class:`.DisplayOpts` instances for every
        overlay.
        """

        for display in self.__displays.values():
            opts     = display.opts
            optProps = opts.getVolumeProps()

            for optProp in optProps:
                if self.syncOverlayVolume:
                    if opts.canBeSyncedToParent(optProp):
                        opts.syncToParent(optProp)
                else:
                    if opts.canBeUnsyncedFromParent(optProp):
                        opts.unsyncFromParent(optProp)


    def __updateBounds(self, *a):
        """Called when the overlay list changes, or when any overlay display
        transform is changed. Updates the :attr:`bounds` property so that it
        is big enough to contain all of the overlays (as defined by their
        :attr:`.DisplayOpts.bounds` properties).
        """

        if len(self.__overlayList) == 0:
            minBounds = [0.0, 0.0, 0.0]
            maxBounds = [0.0, 0.0, 0.0]

        else:
            minBounds = 3 * [ sys.float_info.max]
            maxBounds = 3 * [-sys.float_info.max]

        for ovl in self.__overlayList:

            display = self.__displays[ovl]
            opts    = display.opts
            lo      = opts.bounds.getLo()
            hi      = opts.bounds.getHi()

            for ax in range(3):

                if lo[ax] < minBounds[ax]: minBounds[ax] = lo[ax]
                if hi[ax] > maxBounds[ax]: maxBounds[ax] = hi[ax]

        self.bounds[:] = [minBounds[0], maxBounds[0],
                          minBounds[1], maxBounds[1],
                          minBounds[2], maxBounds[2]]

        # Update the constraints on the location
        # property to be aligned with the new bounds
        with props.suppress(self, 'location'):
            self.location.setLimits(0, self.bounds.xlo, self.bounds.xhi)
            self.location.setLimits(1, self.bounds.ylo, self.bounds.yhi)
            self.location.setLimits(2, self.bounds.zlo, self.bounds.zhi)


    def __locationChanged(self, *a):
        """Called when the :attr:`location` property changes. Propagates
        the new location to the :attr:`worldLocation` property.
        """
        self.__propagateLocation('world')


    def __worldLocationChanged(self, *a):
        """Called when the :attr:`worldLocation` property changes. Propagates
        the new location to the :attr:`location` property.
        """

        self.__propagateLocation('display')


    def __propagateLocation(self, dest):
        """Called by the :meth:`__locationChanged` and
        :meth:`__worldLocationChanged` methods. The ``dest`` argument may be
        either ``'world'`` (the ``worldLocation`` is updated from the
        ``location``), or ``'display'`` (vice-versa).
        """

        if self.displaySpace == 'world':
            if dest == 'world':
                with props.skip(self, 'worldLocation', self.__name):
                    self.worldLocation = self.location
            else:
                with props.skip(self, 'location', self.__name):
                    self.location = self.worldLocation
            return

        if self.displaySpace == 'scaledVoxel':
            ref      = self.getSelectedOverlay()
            srcSpace = 'pixdim'
        else:
            ref      = self.displaySpace
            srcSpace = 'display'

        if ref is None:
            return

        opts = self.getOpts(ref)

        if dest == 'world':
            with props.skip(self, 'location', self.__name):
                self.worldLocation = opts.transformCoords(
                    self.location, srcSpace, 'world')
        else:
            with props.skip(self, 'worldLocation', self.__name):
                self.location = opts.transformCoords(
                    self.worldLocation, 'world', srcSpace)



#!/usr/bin/env python
#
# shopts.py - The SHOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SHOpts` class, a :class:`.VectorOpts`
class for rendering :class:`.Image` instances which contain fibre orientation
distributions (FODs) in the form of spherical harmonic (SH) coefficients.
"""


import os.path       as op

import numpy         as np

import fsleyes_props as props

import                  fsleyes
from . import           vectoropts


SH_COEFFICIENT_TYPE = {
    1   : ('asym', 0),
    9   : ('asym', 2),
    25  : ('asym', 4),
    49  : ('asym', 6),
    81  : ('asym', 8),
    121 : ('asym', 10),
    169 : ('asym', 12),
    225 : ('asym', 14),
    289 : ('asym', 16),
    # 1   : ('sym',  0),  # same as asym, 0
    6   : ('sym',  2),
    15  : ('sym',  4),
    28  : ('sym',  6),
    45  : ('sym',  8),
    66  : ('sym',  10),
    91  : ('sym',  12),
    120 : ('sym',  14),
    153 : ('sym',  16),
}
"""``Image`` files which contain SH coefficients may be symmetric (only
containing coefficients for even spherical functions) or asymmetric
(containing coefficients for odd and even functions). This dictionary provides
mappings from the number coefficients (the volumes contained in the image), to
the file type (either symmetric [``'sym'``] or asymmetric [``'asym'``), and the
maximum SH order that was used in generating the coefficients.
"""


class SHOpts(vectoropts.VectorOpts):
    """The ``SHOpts`` is used for rendering :class:`.Image` instances which
    contain fibre orientation distributions (FODs) in the form of spherical
    harmonic (SH) coefficients. A ``SHOpts`` instance will be used for
    ``Image`` overlays with a :attr:`.Displaty.overlayType` set to ``'sh'``.


    A collection of pre-calculated SH basis function parameters are stored in
    the ``assets/sh/`` directory. Depending on the SH order that was used in
    the fibre orientation, and the desired display resolution (controlled by
    :attr:`shResolution`), a different set of parameters needs to be used.
    The :meth:`getSHParameters` method will load and return the corrrect
    set of parameters.
    """


    shResolution = props.Int(minval=3, maxval=10, default=5)
    """Resolution of the sphere used to display the FODs at each voxel. The
    value is equal to the number of iterations that an isocahedron, starting
    with 12 vertices, is tessellated. The resulting number of vertices is
    as follows:


    ==================== ==================
    Number of iterations Number of vertices
    3                    92
    4                    162
    5                    252
    6                    362
    7                    492
    8                    642
    9                    812
    10                   1002
    ==================== ==================
    """


    shOrder = props.Choice(allowStr=True)
    """Maximum spherical harmonic order to visualise. This is populated in
    :meth:`__init__`.
    """


    size = props.Percentage(minval=0.1, maxval=500, default=100)
    """Display size - this is simply a linear scaling factor. """


    lighting = props.Boolean(default=False)
    """Apply a simple directional lighting model to the FODs. """


    radiusThreshold = props.Real(minval=0.0, maxval=1.0, default=0.05)
    """FODs with a maximum radius that is below this threshold are not shown.
    """


    normalise = props.Boolean(default=False)
    """If ``True``, the size of each FOD is scaled to fit within a
    voxel. Otherwise the size of each FOD is scaled relative to the entire
    image.
    """


    colourMode = props.Choice(('direction', 'radius'))
    """How to colour each FOD. This property is overridden if the
    :attr:`.VectorOpts.colourImage` is set.

      - ``'direction'`` The vertices of an FOD are coloured according to their
                        x/y/z location (see :attr:`xColour`, :attr:`yColour`,
                        and :attr:`zColour`).

      - ``'radius'``    The vertices of an FOD are coloured according to their
                        distance from the FOD centre (see :attr:`colourMap`).
    """


    def __init__(self, *args, **kwargs):

        vectoropts.VectorOpts.__init__(self, *args, **kwargs)

        ncoefs           = self.overlay.shape[3]
        shType, maxOrder = SH_COEFFICIENT_TYPE.get(ncoefs)

        if shType is None:
            raise ValueError('{} does not look like a SH '
                             'image'.format(self.overlay.name))

        self.__maxOrder = maxOrder
        self.__shType   = shType

        # If this Opts instance has a parent,
        # the shOrder choices will be inherited
        if self.getParent() is None:

            if   shType == 'sym':  vizOrders = range(0, self.__maxOrder + 1, 2)
            elif shType == 'asym': vizOrders = range(0, self.__maxOrder + 1)

            self.getProp('shOrder').setChoices(list(vizOrders), instance=self)
            self.shOrder = vizOrders[-1]


    @property
    def shType(self):
        """Returns either ``'sym'`` or ``'asym'``, depending on the type
        of the SH coefficients contained in the file.
        """
        return self.__shType


    @property
    def maxOrder(self):
        """Returns the maximum SH order that was used to generate the
        coefficients of the SH image.
        """
        return self.__maxOrder


    def getSHParameters(self):
        """Load and return a ``numpy`` array containing pre-calculated SH
        function parameters for the curert maximum SH order and display
        resolution. The returned array has the shape ``(N, C)``, where ``N``
        is the number of vertices used to represent each FOD, and ``C`` is
        the number of SH coefficients.
        """

        # TODO Adjust matrix if shOrder is
        #      less than its maximum possible
        #      value for this image.
        #
        #      Also, calculate the normal vectors.

        resolution = self.shResolution
        ncoefs     = self.overlay.shape[3]
        order      = self.shOrder
        ftype, _   = SH_COEFFICIENT_TYPE[ncoefs]
        fname     = op.join(
            fsleyes.assetDir,
            'sh',
            '{}_coef_{}_{}.txt'.format(ftype, resolution, order))

        params = np.loadtxt(fname)

        if len(params.shape) == 1:
            params = params.reshape((-1, 1))

        return params


    def getVertices(self):
        """Loads and returns a ``numpy`` array of shape ``(N, 3)``, containing
        ``N`` vertices of a tessellated sphere.
        """
        fname = op.join(
            fsleyes.assetDir,
            'sh',
            'vert_{}.txt'.format(self.shResolution))

        return np.loadtxt(fname)


    def getIndices(self):
        """Loads and returns a 1D ``numpy`` array, containing indices into
        the vertex array, specifying the order in which they are to be drawn
        as triangles.
        """
        fname = op.join(
            fsleyes.assetDir,
            'sh',
            'face_{}.txt'.format(self.shResolution))

        return np.loadtxt(fname).flatten()
